//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%
#include <XSPlot/Plot/CreateBinnedPlotGroups.h>
#include <XSPlot/Plot/PlotGroup.h>
#include <XSPlot/Plot/PlotSettings.h>
#include <XSPlot/Plot/PlotStyle.h>
#include <XSPlot/Plot/PlotTypes.h>
#include <XSUtil/Numerics/Numerics.h>
#include <XSstreams.h>
#include <XSUtil/Utils/XSstream.h>

// PlotData
#include <XSPlot/Commands/PlotData.h>


// Class PlotData 

PlotData::PlotData (const string& name, bool isCounts)
  : PlotCommand(name, new CreateBinnedPlotGroups((isCounts ? COUNTS : RATE))),
   m_isCounts(isCounts)
{
  doesTakePlotGroupNumbers(true);
  isDataRequired(true);
  isActiveModelRequired(false);
  isDivisibleByArea(true);
  addCompLevel(2);
  isBackgroundApplicable(true);
  isLineIDApplicable(true);
  isPlotGroupDonor(true);

  PlotAttributes styles;
  styles.symbolStyle = PlotCommand::standardDataStyle();
  styles.lineStyle = PlotStyle::NONE;
  setStyleMap(PlotStyle::DATA, styles);
  styles.symbolStyle = PlotStyle::ASTERISK;
  styles.lineStyle = PlotStyle::NONE; 
  setStyleMap(PlotStyle::BACKGROUND, styles); 
  styles.symbolStyle = PlotStyle::BLANK;
  styles.lineStyle = PlotCommand::standardModelStyle();
  styles.lineStep = true;
  styles.lineWidth = 2;
  setStyleMap(PlotStyle::MODEL, styles);
  styles.lineStep = false;
  styles.lineStyle = PlotStyle::DOTTED;
  setStyleMap(PlotStyle::SOURCES, styles);
}


PlotData::~PlotData()
{
}

void PlotData::manipulate (std::vector<PlotGroup*>& plotGroups, const PlotSettings& settings)
{
   for (size_t i=0; i<plotGroups.size(); ++i)
   {
      PlotGroup* pg = plotGroups[i];
      const size_t nPts = pg->m_n;
      std::vector<Real>& X      = pg->m_xAxis.data;
      std::vector<Real>& dX     = pg->m_xAxis.errors[0];
      const Real HZTOERG = Numerics::KEVTOERG/Numerics::KEVTOHZ;
      std::vector<Real> multiplier(nPts, 1.0);

      // strip the l off the front of the command name since the multiplier factors will
      // be the same for commands with and without the initial l.
      string cmdStr = cmdName();
      if ( cmdStr.substr(0,1) == "l" ) cmdStr = cmdName().substr(1);

      // Data vector is in counts/s
      if (settings.xOption() == ENERGY) {
	if (settings.getUnitID(ENERGY) == "Hz") {
	  // if energy setting is Hz we don't want to use that to multiply by
	  // ebergy but we want to end up in ergs
	  if ( cmdStr == "edata" ) {
	    // Want to ultimately display in ergs/s/Hz, need
	    // to convert from Hz*counts/s/Hz by applying a HZTOERG factor
	    // Note that the /Hz factor is handled in CreateBinnedPlotGroups
	    for (size_t k=0; k<nPts; ++k)
	      multiplier[k] = HZTOERG*sqrt((X[k]-dX[k])*(X[k]+dX[k]));
	  } else if ( cmdStr == "eedata" ) {
	    // Want ergs/s
	    // X is in Hz.
	    for (size_t k=0; k<nPts; ++k)
	      multiplier[k] = HZTOERG*(X[k]-dX[k])*(X[k]+dX[k]);
	  }
	} else {
	  if ( cmdStr == "edata" ) {
	    // Want <units>(counts/s/<units>)
	    // X is already in the selected <units> of energy so
	    // no need for conversion.
	    for (size_t k=0; k<nPts; ++k) multiplier[k] = sqrt((X[k]-dX[k])*(X[k]+dX[k]));
	  } else if ( cmdStr == "eedata" ) {
	    // Want <units>^2(counts//s/<units>)
	    // X is in <units> (keV,MeV,GeV).
	    for (size_t k=0; k<nPts; ++k) multiplier[k] = (X[k]-dX[k])*(X[k]+dX[k]);
	  }
	}
	// end if ENERGY mode
      } else {
	// WAVELENGTH mode
	// For the wavelength case we still want to multiply by E or E^2
	// though X and dX are lengths.
	const Real unitFactor = settings.getUnitFactor(WAVELENGTH);
	if (settings.isWavePerHz()) {
	  if ( cmdStr == "edata" ) {
	    // eventually we want erg/s/Hz
	    // Note that the /Hz factor is handled in CreateBinnedPlotGroups
	    const Real conversion = Numerics::KEVTOA*Numerics::KEVTOERG*unitFactor;
	    for (size_t k=0; k<nPts; ++k) multiplier[k] = conversion/sqrt(X[k]*X[k]-dX[k]*dX[k]);
	  } else if ( cmdStr == "eedata" ) {
	    // eventually we want ergs/s
	    const Real conversion = Numerics::KEVTOERG*Numerics::KEVTOA*unitFactor;
	    for (size_t k=0; k<nPts; ++k) {
	      multiplier[k] = conversion*conversion/(X[k]*X[k]-dX[k]*dX[k]);
	      // now convert multiplier from ergs^2 to erg-Hz
	      multiplier[k] /= HZTOERG;
	    }
	  }
	  // end if perHz
	} else {
	  // per <length> units
	  if ( cmdStr == "edata" ) {
	    // Want E in ergs.  So multiply by conversion
	    // = hc(in keV-A)*(ergs/keV)*(<units>/Angstrom)
	    // = KEVTOA*KEVTOERG*unitFactor.
	    Real conversion = Numerics::KEVTOA*Numerics::KEVTOERG*unitFactor;
	    for (size_t k=0; k<nPts; ++k) {
	      multiplier[k] = conversion/sqrt(X[k]*X[k]-dX[k]*dX[k]);
	    }
	  } else if ( cmdStr == "eedata" ) {
	    // Want ergs^2/s/<length>
	    const Real conversion = Numerics::KEVTOA*Numerics::KEVTOERG*unitFactor;
	    for (size_t k=0; k<nPts; ++k) {
	      Real E = conversion*X[k]/(X[k]*X[k]-dX[k]*dX[k]);
	      multiplier[k] = E*E;
	    }
	  }
	} // end if per length
      } // end if wavelength mode

      // now do all the multiplications
      std::vector<Real>& yData  = pg->m_yData.data;
      std::vector<Real>& yError = pg->m_yData.errors[0];
      std::vector<Real>& background = pg->m_background.data;
      std::vector<Real>& backgroundError = pg->m_background.errors[0];

      for (size_t k=0; k<nPts; ++k) {
	if (multiplier[k] != 1.0) {
	  const Real mult = multiplier[k];
	  yData[k] *= mult;
	  yError[k] *= mult;
	  background[k] *=mult;
	  backgroundError[k] *=mult;
	}
      }

      for (size_t i=0; i<pg->m_model.size(); i++) {
	std::vector<Real>& foldmodel = pg->m_model[i].data;
	for (size_t k=0; k<nPts; ++k) {
	  if (multiplier[k] != 1.0) foldmodel[k] *= multiplier[k];
	}
      }

      for (size_t i=0; i<pg->m_sources.size(); i++) {
	PlotVectorList::iterator itSource = pg->m_sources[i].begin();
	PlotVectorList::iterator itEnd = pg->m_sources[i].end();
	while (itSource != itEnd) {
	  for (size_t k=0; k<nPts; k++) itSource->data[k] *= multiplier[k];
	  ++itSource;
	}
      }

      for (size_t i=0; i<pg->m_auxData.size(); i++) {
	std::vector<Real>& auxData = pg->m_auxData[i].data;
	std::vector<Real>& auxDataError = pg->m_auxData[i].errors[0];
	for (size_t k=0; k<nPts; ++k) {
	  if (multiplier[k] != 1.0) {
	    const Real mult = multiplier[k];
	    auxData[k] *= mult;
	    auxDataError[k] *= mult;
	  }
	}
      }

   } // end plot groups loop
}

void PlotData::setRanges (std::vector<PlotGroup*>& plotGroups, const PlotSettings& settings)
{
   Real xlow=0.0, xhigh=0.0;
   PlotGroupCreator::determineXRange(plotGroups,&xlow,&xhigh);

   PlotRange& rangeVals = rangeSettings();
   rangeVals.ranges.x1 = xlow;
   rangeVals.ranges.x2 = xhigh;
   const string cmdStr = cmdName();

   if (cmdStr.substr(0,1) == "l")
   {
      // Y-scale will be log regardless of setplot setting.
      rangeVals.yScaleType = PlotStyle::LOG;
   }
   else
      rangeVals.yScaleType = settings.yLog() ? PlotStyle::LOG : 
                PlotStyle::LINEAR;
   if (settings.xOption() == CHANNELS)
      rangeVals.xScaleType = PlotStyle::LINEAR;
   else 
      rangeVals.xScaleType = settings.xLog() ? PlotStyle::LOG :
                PlotStyle::LINEAR;


   Real ylow=0.0, yhigh=0.0;
   PlotGroupCreator::determineYRange(plotGroups,&ylow,&yhigh,settings.getShowBackground());

   if (ylow == yhigh)
   {
      // Never want to set the same values for ylow and yhigh.
      if (ylow == 0.0)
      {
         // Just give an arbitrary value to yhigh, as long
         //  as it's > 0.0.
         yhigh = 1.0;
      }
      else
      {
         yhigh *= 2.0;
         ylow /= 2.0;
      }
   }

   if (rangeVals.yScaleType == PlotStyle::LOG) {
     ylow = std::max(ylow, yhigh*1.0e-8);
     rangeVals.ranges.y1 = ylow;
   } else {
     // ylow can go negative due to lower error bars.  That's
     // the only time we'll set the low range to differ from 0.0.
     rangeVals.ranges.y1 = std::min(0.0, ylow);
   }
   rangeVals.ranges.y2 = yhigh;
   
   rangeVals.useRangeFlags = PlotStyle::XMIN + PlotStyle::XMAX
     + PlotStyle::YMIN + PlotStyle::YMAX;
}

void PlotData::makeLabels (const PlotSettings& settings, StandardLabels& labels) const
{
   labels.title = activeModelStatus() ? settings.lookupLabelName("t:data:foldmodel") :
                        settings.lookupLabelName("t:data");

   // strip the l off the front of the command name since the labels will be the same for
   // commands with and without the initial l. this means we only need entries for the non-l
   // cases in plotLabels.dat
   string cmdStr = cmdName();
   if ( cmdStr.substr(0,1) == "l" ) cmdStr = cmdName().substr(1);

   string xOption = settings.makeXOptionDependentLabel(labels.x).substr(2);
   string yKey("y:"+cmdStr+":");
   // for ENERGY using Hz units use the same y-label format as for
   // WAVELENGTH in perHz mode except for the data and ldata cases
   if ( cmdStr == "data" || cmdStr == "counts") {
     yKey += xOption;
   } else {
     if ( xOption == "channel" || ((xOption == "energy")&&(settings.getUnitID(ENERGY) != "Hz")) ||
	  ((xOption == "wave")&&(!settings.isWavePerHz())) ) {
       yKey += xOption;
     } else {
       yKey += "wave:perHz";
     }
   }

   string yLabel = settings.lookupLabelName(yKey);
   settings.insertUnitLabel(yLabel, true);

   if (settings.divideByArea()) {
     string inverseArea = settings.lookupLabelName("y:suffix:area");
     yLabel += " ";
     yLabel += inverseArea;
   } else {
     if ( settings.divideByAreascaleOnly() ) {
       string inverseArea = settings.lookupLabelName("y:suffix:areascale");
       yLabel += " ";
       yLabel += inverseArea;
     }
   }
   labels.y = yLabel;

}

// Additional Declarations
