#ifndef RANDOMIZERPOLICIES_H
#define RANDOMIZERPOLICIES_H 1

#include <xsTypes.h>

class Fit;
class ChainManager;

//These classes don't need to do anything other than be types.
class RandomizerCovarFile {};
class RandomizerCmdLine {};
class RandomizerDeltas {};
class RandomizerLimits {};

// Source traits class, though additional function runInitialization() actually
// makes this a traits/policy combo.

// Note that due to the structural similarities in the specialized classes,
// a slightly better design would move this commonality into the generic
// RandomizerSourceInfo<T> class.  However a problem with Sun CC 
// (naturally) prevents doing this.  The sticking point is that it's
// unable to properly deal with individually specializing the
// static const name data member.  This was verified on CC 5.3
// and 5.4, and likely exists on other versions too. 

template <class Source>
class RandomizerSourceInfo;

template <>
class RandomizerSourceInfo<Fit>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_eigenValues;}
      const RealArray& eigenValues() const {return m_eigenValues;}
      const RealArray& eigenVectors() const {return m_eigenVectors;}
      const RealArray& covariance(const Fit* fit);
      void loadInitialization(const string& initStr) {}
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:
      RealArray m_eigenValues;
      RealArray m_eigenVectors;
      RealArray m_covariance;
};

template <>
class RandomizerSourceInfo<ChainManager>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_eigenValues;}
      const RealArray& eigenValues() const {return m_eigenValues;}
      const RealArray& eigenVectors() const {return m_eigenVectors;}
      const RealArray& covariance(const Fit* fit);
      void loadInitialization(const string& initStr) {}
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:
      RealArray m_eigenValues;
      RealArray m_eigenVectors;
      RealArray m_covariance;
};

template <>
class RandomizerSourceInfo<RandomizerCovarFile>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_eigenValues;}
      const RealArray& eigenValues() const {return m_eigenValues;}
      const RealArray& eigenVectors() const {return m_eigenVectors;}
      const RealArray& covariance(const Fit* fit) const {return m_covariance;}

      // This function will enforce the constraint that the matrix
      // stored in the file is a square matrix (To be more 
      // precise, it enforces that the lower half belongs to a
      // square.  Terms above the diagonal are ignored.), and that
      // the number of pars matches the current number of 
      // variable fit pars. 
      void loadInitialization(const string& initStr);
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:
      void readFromFile(const string& fileName);

      RealArray m_eigenValues;
      RealArray m_eigenVectors;
      RealArray m_covariance;
};

template <>
class RandomizerSourceInfo<RandomizerCmdLine>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_eigenValues;}
      const RealArray& eigenValues() const {return m_eigenValues;}
      const RealArray& eigenVectors() const {return m_eigenVectors;}
      const RealArray& covariance(const Fit* fit) const {return m_covariance;}

      void loadInitialization(const string& initStr);
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:
      int parseCmdLine(const string& lineArgs);

      RealArray m_eigenValues;
      RealArray m_eigenVectors;
      RealArray m_covariance;
      bool m_onlyDiag;
};


template <>
class RandomizerSourceInfo<RandomizerDeltas>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_eigenValues;}
      const RealArray& eigenValues() const {return m_eigenValues;}
      const RealArray& eigenVectors() const {return m_eigenVectors;}
      const RealArray& covariance(const Fit* fit) const {return m_covariance;}

      void loadInitialization(const string& initStr);
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:

      RealArray m_eigenValues;
      RealArray m_eigenVectors;
      RealArray m_covariance;
      Real m_multFactor;
};

template <>
class RandomizerSourceInfo<RandomizerLimits>
{
   public:
      static const string name;

      const RealArray& scalingFactors() const {return m_limitRanges;}
      const RealArray& limitMinimums() const {return m_limitMinimums;}
      const RealArray& limitRanges() const {return m_limitRanges;}
      // this is a bit ugly and is only included for parallelism - RandomizerLimits doesn't use covariance
      const RealArray& covariance(const Fit* fit) const {return {};}

      void loadInitialization(const string& initStr);
      void runInitialization(const Fit* fit);
      void applyRandomization(const RealArray& randVals, RealArray& parameterValues);

   private:

      // limits for each parameter run from m_limitMinimums to m_limitMinimums+m_limitRanges
      RealArray m_limitMinimums;
      RealArray m_limitRanges;
};


class RandomizerGaussDist
{
   public:
      static const string name;

      static void getValuesFromDistribution(const RealArray& scaling, RealArray& randVals);
};

class RandomizerCauchyDist
{
   public:
      static const string name;

      static void getValuesFromDistribution(const RealArray& scaling, RealArray& randVals);
};

class RandomizerUniformDist
{
   public:
      static const string name;

      static void getValuesFromDistribution(const RealArray& scaling, RealArray& randVals);
};

// Useful routine for the various Randomizer options

void randomizeUsingEigenvectors(const RealArray& randVals, const RealArray& eVectors,RealArray& parameterValues);

#endif
