// Voigt profile emission line
//  Parameters are :
//    0:   ecenter, line center (keV)
//    1:   sigma from Gaussian  (keV)
//    2:   gamma from Lorentzian  (keV)

// Calculation via the Faddeeva function w(z) = exp(-z^2) erfc(-iz)
// in XSUtil/Numerics/Faddeeva.{cxx,h} from http://ab-initio.mit.edu/Faddeeva
//    Voigt(x,sigma,gamma) = Re[w(z)] / sigma*sqrt(2pi)
// where z = (x + i*gamma)/sigma*sqrt(2) and x = E-ecenter.

#include <XSFunctions/functionMap.h>
#include <XSFunctions/Utilities/FunctionUtility.h>
#include <xsTypes.h>
#include <XSUtil/Numerics/AdaptiveIntegrate.h>
#include <XSUtil/Numerics/BinarySearch.h>
#include <XSUtil/Numerics/Faddeeva.h>
#include <cmath>
#include <complex>

using namespace Numerics;

RealArray voigtLineIntegrand(const RealArray& x, void *p);

void voigtLine(const RealArray& energyArray, const RealArray& params, 
	       int spectrumNumber, RealArray& fluxArray, 
	       RealArray& fluxErrArray, const string& initString)
{

  Real center = params[0];
  Real sigma = params[1];
  Real gamma = params[2];

  // trap special cases. if sigma = 0 then the line shape is lorentzian and if gamma = 0
  // then it is gaussian

  if ( sigma <= 0.0 ) {
    RealArray lparams(2);
    lparams[0] = center;
    lparams[1] = gamma;
    lorentzianLine(energyArray, lparams, spectrumNumber, fluxArray, fluxErrArray, initString);
    return;
  }
  if ( gamma <= 0.0 ) {
    RealArray gparams(2);
    gparams[0] = center;
    gparams[1] = sigma;
    gaussianLine(energyArray, gparams, spectrumNumber, fluxArray, fluxErrArray, initString);
    return;
  }

  size_t nE = energyArray.size();
  fluxArray.resize(nE-1);
  fluxErrArray.resize(0);
  fluxArray = 0.0;

  // set the number of fwhm over which to evaluate the model. By
  // default this is 10 but the user can over ride using
  // xset VOIGTCRITFWHM value

  string pname = "VOIGTCRITFWHM";
  string pvalue = FunctionUtility::getModelString(pname);
  Real crtFwhm = 10;
  if (pvalue.length() && pvalue != FunctionUtility::NOT_A_KEY()) {
    std::istringstream iss(pvalue);
    Real tmpVal = 10.0;
    if ( !(iss >> tmpVal) || !iss.eof() ) {
      std::ostringstream oss;
      oss << "Failed to read value from VOIGTCRITFWHM - assuming " << crtFwhm << "\n";
      FunctionUtility::xsWrite(oss.str(),10);
    } else {
      crtFwhm = tmpVal;
    }
  }

  // evaluate the line over +/- crtFwhm * fwhm where the fwhm of the
  // Voigt line profile is estimated by (Kielkopf 1973)
  // fwhm = 0.5346*(2*gamma) + sqrt(0.2166*(2*gamma)^2 + (2*sigma)^2*2*ln(2))

  const Real fwhm = 0.5346*(2.0*gamma)
    + sqrt(0.2166*(2.0*gamma)*(2.0*gamma) + (2.0*sigma)*(2.0*sigma)*2.0*log(2.0));
  Real emin = center - crtFwhm * fwhm;
  Real emax = center + crtFwhm * fwhm;

  Real efirst = energyArray[0];
  Real elast = energyArray[nE-1];

  // If the line calculation energies lie outside the energy range then do nothing
  if ( emax < efirst || emin > elast ) return;

  // Find the first and last bins on which we need to calculate the line

  int ifirst(0), ilast(nE-2);
  if ( emin >= efirst ) ifirst = BinarySearch(energyArray, emin);
  if ( emax <= elast ) ilast =  BinarySearch(energyArray, emax);

  // if the width is zero the entire line is in one bin
  if ( fwhm == 0 ) {
    fluxArray[ifirst] = 1.0;
    return;
  }

  // integrate the line over the energy bins of interest to a precision of Precision
  // in each bin

  Real Precision = 1.0e-6;
  RealArray p(params);

  int ModelEvaluations(0);
  for (int i=ifirst; i<=ilast; i++) {
    Real Integral, IntegralError;
    int iter = AdaptiveIntegrate<voigtLineIntegrand>(energyArray[i], energyArray[i+1],
						     &p, Precision, Integral, IntegralError);
    fluxArray[i] = Integral;
    // basic Gauss-Kronrod integration uses 15 model evaluations and each iteration of the
    // adaptive integration does two additional G-K integrations.
    ModelEvaluations += 15 + iter*30;
  }

  const string dbKeyword("VOIGT_MODEL_EVALS");
  FunctionUtility::loadDbValue(dbKeyword, (double)ModelEvaluations);

  return;
}

// evaluate the Voigt line integrand with p used to pass in the parameter values

RealArray voigtLineIntegrand(const RealArray& x, void *p)
{
  RealArray OutArray(x.size());

  RealArray params(3);
  params = *(static_cast<RealArray*>(p));

  Real center = params[0];
  Real sigma = params[1];
  Real gamma = params[2];

  Real vnorm = 1.0/sigma/sqrt(2*M_PI);

  Real znorm = 1.0/sigma/sqrt(2.0);
  for (size_t i = 0; i<x.size(); i++) {
    std::complex<double> z(x[i]-center,gamma);
    z *= znorm;
    std::complex<double> wofz = Faddeeva::w(z);
    OutArray[i] = vnorm * wofz.real();
  }

  return OutArray;
}
